import logging
import sys
import asyncio
import signal
import time
from functools import partial

import rapidjson
from humps import camelize
import structlog
from structlog import get_logger
from tornado import ioloop, httpserver
from tornado.options import define, options


class Server:
    def __init__(self, create_dependencies, config_path):
        self._config_factory = ConfigFactory(config_path)
        self._create_dependencies = create_dependencies

    def run(self):
        config = self._get_config()
        log = create_logger(config.get("env") == "production", config.get("log"))
        dependencies = self._create_dependencies(config, log)

        lifecycle_log = log.bind(is_lifecycle_log=True, is_app_log=False)

        app = dependencies.app
        log = dependencies.log
        lifecycle = dependencies.lifecycle

        create_server(config, app, lifecycle_log)

        lifecycle_log.info("Initialize: Start")
        lifecycle_log.info(f"Server: http://localhost:{options.port}")
        io_loop = ioloop.IOLoop.instance()

        try:
            lifecycle.on_start()
        except BaseException as err:  # pylint: disable=broad-except
            lifecycle_log.critical("On Start: Fail", exc_info=err, stack_info=True)
            sys.exit(2)

        io_loop.start()
        lifecycle_log.info("Startup: Success")

    def update_config_factory(self, configure):
        self._config_factory = configure(self._config_factory)

    def _get_config(self):
        return self._config_factory.create()


class ConfigFactory:
    def __init__(self, config_path):  # pylint: disable=unused-argument
        self._config = {}

    def update(self, key, value):
        self._config[key] = value

    def create(self):
        return self._config


def create_logger(is_prod, log_config):
    def log_serializer(data, **kw_args):  # pylint: disable=unused-argument
        if log_config.get("camelize"):
            return rapidjson.dumps(camelize(data))
        return rapidjson.dumps(data)

    processors = [
        structlog.stdlib.add_log_level,
        structlog.processors.format_exc_info,
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.dev.ConsoleRenderer(),
    ]

    if is_prod:
        processors = [
            structlog.stdlib.filter_by_level,
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_log_level_number,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.TimeStamper(utc=True, key="time"),
            structlog.processors.JSONRenderer(serializer=log_serializer),
        ]

    logging.basicConfig(format="%(message)s", level=log_config.get("level").upper())

    structlog.configure(
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
        processors=processors,
    )

    log = get_logger()
    log_props = get_log_props(is_prod, log_config)
    return log.bind(**log_props)


def get_log_props(is_prod, log_config):
    if not is_prod:
        return {}
    return {
        "@service": log_config.get("service"),
        "@system": log_config.get("system"),
        "@env": log_config.get("env"),
        "name": log_config.get("name"),
        "version": log_config.get("version"),
        "is_app_log": True,
    }


def create_server(config, app, log):
    default_port = config.get("port")
    default_shutdown_delay = config.get("shutdown_delay")
    default_debug = config.get("env") == "development"

    define("port", default=default_port, help="run on the given port", type=int)
    define(
        "shutdown_delay",
        default=default_shutdown_delay,
        help="time to wait for shutdown",
        type=int,
    )
    define("debug", default=default_debug, help="run in debug mode", type=int)

    server = httpserver.HTTPServer(app)
    server.listen(options.port)

    signal.signal(signal.SIGTERM, partial(handle_signal, server, log))
    signal.signal(signal.SIGINT, partial(handle_signal, server, log))

    return server


def handle_signal(server, log, sig, frame):  # pylint: disable=unused-argument
    log.info("Signal: Interrupt")

    io_loop = ioloop.IOLoop.instance()

    def stop_loop(server, deadline):
        now = time.time()

        tasks = [
            t
            for t in asyncio.all_tasks()
            if t is not asyncio.current_task() and not t.done()
        ]

        if now < deadline and len(tasks) > 0:
            log.info("Shutdown: Wait", pending_tasks=len(tasks))
            io_loop.add_timeout(now + 1, stop_loop, server, deadline)
            return

        pending_connection = len(
            server._connections  # pylint: disable=protected-access
        )
        if now < deadline and pending_connection > 0:
            log.info("Shutdown: Wait", connections=pending_connection)
            io_loop.add_timeout(now + 1, stop_loop, server, deadline)
        else:
            log.info("Server: Close", connections=pending_connection)
            io_loop.stop()
            log.info("Shutdown: Success")

    def shutdown():
        try:
            log.info("Shutdown: Start", delay=options.shutdown_delay)
            deadline = time.time() + options.shutdown_delay
            stop_loop(server, deadline)
        except BaseException as err:  # pylint: disable=broad-except
            log.critical("Shutdown: Fail", exc_info=err, stack_info=True)
            sys.exit(1)

    io_loop.add_callback_from_signal(shutdown)



from os import environ


def configure(config_factory):
    log_config(config_factory)
    server_config(config_factory)
    app_config(config_factory)
    return config_factory


def server_config(config_factory):
    env = environ.get("PYAPP_ENV") or "development"
    is_dev = env == "development"
    config_factory.update("env", env)
    config_factory.update("debug", environ.get("PYAPP_DEBUG") or is_dev)
    config_factory.update("port", environ.get("PORT") or 9000)
    config_factory.update("shutdown_delay", environ.get("PYAPP_SHUTDOWN_DELAY") or 3)


def app_config(config_factory):  # pylint: disable=unused-argument
    pass


def log_config(config_factory):
    config = {
        "camelize": environ.get("LOG_CAMELIZE") == "true",
        "level": environ.get("LOG_LEVEL") or "info",
        "service": environ.get("LOG_SERVICE"),
        "system": environ.get("LOG_SYSTEM"),
        "env": environ.get("LOG_ENV"),
        "name": environ.get("LOG_NAME"),
        "version": environ.get("LOG_VERSION"),
    }
    config_factory.update("log", config)



from os import environ, path

from .server import Server
from .config import configure


def boot(create_dependencies):
    config_path = environ.get("PYAPP_CONFIG_PATH") or path.realpath(
        path.join(path.dirname(__file__), "..", "..", "config")
    )
    server = Server(create_dependencies, config_path)
    server.update_config_factory(configure)
    server.run()



import asyncio
import platform

from . import create_dependencies, boot


def main():
    boot(create_dependencies)


if __name__ == "__main__":
    if platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    main()



from tornado.web import Application, RequestHandler


class App(Application):
    def __init__(self, app_handlers, app_settings):
        handlers = app_handlers.get()
        settings = app_settings.get()
        super().__init__(handlers=handlers, **settings)


class AppHandlers:
    def __init__(self, flags, log):
        opts = dict(flags=flags, log=log)
        self._handlers = [
            (r"/health", HealthHandler, opts),
        ]

    def get(self):
        return self._handlers


class AppSettings:
    def __init__(self, log, flags):
        self._log = log.bind(is_request_log=True, is_app_log=False)
        self._settings = dict(debug=flags["debug"], log_function=self._log_function)

    def get(self):
        return self._settings

    def _log_function(self, handler):
        if handler.get_status() < 400:
            log_method = self._log.info
        elif handler.get_status() < 500:
            log_method = self._log.warning
        else:
            log_method = self._log.error

        request_time = 1000.0 * handler.request.request_time()

        log_method(
            handler._request_summary(),  # pylint: disable=protected-access
            req_status=handler.get_status(),
            req_time=request_time,
        )


class ExceptionHandler:
    def log_exception(
        self, exc_type, exc_info, trace_back
    ):  # pylint: disable=unused-argument
        if getattr(self, "_logger"):
            self._logger.error(  # pylint: disable=no-member
                "Request: Error", exc_info=exc_info, stack_info=True
            )


class HealthHandler(ExceptionHandler, RequestHandler):
    def initialize(self, flags, log):  # pylint: disable=unused-argument
        self._logger = log  # pylint: disable=attribute-defined-outside-init

    def data_received(self, chunk):
        pass

    def get(self):
        self.write({"healthy": True})
        self.flush()



from pinject import BindingSpec, new_object_graph, copy_args_to_public_fields

from .app import App, AppHandlers, AppSettings
from .lifecycle import Lifecycle


def create_dependencies(config, log):
    flags = {"debug": config.get("debug")}

    class FlagsBindingSpec(BindingSpec):
        def configure(self, bind):
            bind("flags", to_instance=flags)

    class LogBindingSpec(BindingSpec):
        def configure(self, bind):
            bind("log", to_instance=log)

    binding_specs = [FlagsBindingSpec(), LogBindingSpec()]

    classes = [
        App,
        AppHandlers,
        AppSettings,
        Lifecycle,
    ]

    obj_graph = new_object_graph(
        modules=None, classes=classes, binding_specs=binding_specs
    )

    return obj_graph.provide(Dependencies)


class Dependencies:
    @copy_args_to_public_fields
    def __init__(self, app, lifecycle, log):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['makenew_python_app', 'makenew_python_app.server']

package_data = \
{'': ['*']}

install_requires = \
['pinject>=0.14.1,<0.15.0',
 'pyhumps>=3.0.2,<4.0.0',
 'python-rapidjson>=1.0.0,<2.0.0',
 'structlog>=21.1.0,<22.0.0',
 'tornado>=6.0.4,<7.0.0']

setup_kwargs = {
    'name': 'makenew-python-app',
    'version': '2.13.0',
    'description': 'Package skeleton for a Python app.',
    'long_description': 'Python App Skeleton\n===================\n\n|PyPI| |GitHub Actions|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/makenew-python-app.svg\n   :target: https://pypi.python.org/pypi/makenew-python-app\n   :alt: PyPI\n.. |GitHub Actions| image:: https://github.com/makenew/python-app/workflows/main/badge.svg\n   :target: https://github.com/makenew/python-app/actions\n   :alt: GitHub Actions\n\nPackage skeleton for a Python app.\n\nDescription\n-----------\n\nBootstrap a new Python_ app in less than a minute.\n\n.. _Python: https://www.python.org/\n\nFeatures\n~~~~~~~~\n\n- Publishing to PyPI_.\n- Secure dependency management with Poetry_.\n- Multi-stage Docker_ builds for optimized production images.\n- Images tagged using package version and commit checksum.\n- Images pushed to `GitHub Container Registry`_.\n- Standardized JSON logging with structlog_.\n- Centralized dependency injection with pinject_.\n- Tornado_ with asyncio.\n- Linting with Pylint_.\n- Uncompromising code formatting with Black_.\n- pytest_ helps you write better programs.\n- Code coverage reporting with Codecov_.\n- Continuous unit and smoke testing and deployment with `GitHub Actions`_.\n- `Keep a CHANGELOG`_.\n- Consistent coding with EditorConfig_.\n- Badges from Shields.io_.\n\n.. _Black: https://black.readthedocs.io/en/stable/\n.. _Codecov: https://codecov.io/\n.. _Docker: https://www.docker.com/\n.. _EditorConfig: https://editorconfig.org/\n.. _GitHub Actions: https://github.com/features/actions\n.. _GitHub Container Registry: https://github.com/features/packages\n.. _Keep a CHANGELOG: https://keepachangelog.com/\n.. _PyPI: https://pypi.python.org/pypi\n.. _Pylint: https://www.pylint.org/\n.. _Shields.io: https://shields.io/\n.. _Tornado: https://www.tornadoweb.org/\n.. _pinject: https://pypi.org/project/pinject/\n.. _pytest: https://docs.pytest.org/\n.. _structlog: http://www.structlog.org/\n\nBootstrapping a New Project\n~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n1. Create an empty (**non-initialized**) repository on GitHub.\n2. Clone the master branch of this repository with\n\n   ::\n\n       $ git clone --single-branch https://github.com/makenew/python-app.git new-python-app\n       $ cd new-python-app\n\n   Optionally, reset to the latest\n   `release <https://github.com/makenew/python-app/releases>`__ with\n\n   ::\n\n       $ git reset --hard v1.2.0\n\n3. Run\n\n   ::\n\n       $ ./makenew.sh\n\n   This will replace the boilerplate, delete itself,\n   remove the git remote, remove upstream tags,\n   and stage changes for commit.\n\n4. Create the required GitHub repository secrets\n5. Review, commit, and push the changes to GitHub with\n\n   ::\n\n     $ git diff --cached\n     $ git commit -m "Replace makenew boilerplate"\n     $ git remote add origin git@github.com:<user>/<new-python-app>.git\n     $ git push -u origin master\n\n6. Ensure the GitHub action passes,\n   then publish the initial version of the package with\n\n   ::\n\n     $ poetry install\n     $ poetry run bump2version patch\n     $ git push\n     $ git push --tags\n\nUpdating\n~~~~~~~~\n\nIf you want to pull in future updates from this skeleton,\nyou can fetch and merge in changes from this repository.\n\nAdd this as a new remote with\n\n::\n\n    $ git remote rename origin upstream\n\nand then configure your ``origin`` branch as normal.\n\nOtherwise, add this as a new remote with\n\n::\n\n    $ git remote add upstream git@github.com:makenew/python-app.git\n\nYou can then fetch and merge changes with\n\n::\n\n    $ git fetch --no-tags upstream\n    $ git merge upstream/master\n\nChangelog\n^^^^^^^^^\n\nNote that ``CHANGELOG.md`` is just a template for this skeleton. The\nactual changes for this project are documented in the commit history and\nsummarized under\n`Releases <https://github.com/makenew/python-app/releases>`__.\n\nUsage\n-----\n\nDocker container\n~~~~~~~~~~~~~~~~\n\nThe service is distributed as a Docker container on GitHub Container Registry.\nTo run locally, add configuration to `config/local.json`,\nthen pull and run the image with\n\n::\n\n    $ docker run --read-only --init --publish 8080:8080 \\\n      --volume "$(pwd)/config/local.json:/usr/src/app/config/local.json" \\\n      ghcr.io/makenew/python-app\n\nInstallation\n------------\n\nThis package is also registered on the `Python Package Index (PyPI)`_\nas makenew_python_app_.\n\nInstall it with\n\n::\n\n    $ poetry add makenew_python_app\n\n.. _makenew_python_app: https://pypi.python.org/pypi/makenew-python-app\n.. _Python Package Index (PyPI): https://pypi.python.org/\n\nDevelopment and Testing\n-----------------------\n\nQuickstart\n~~~~~~~~~~\n\n::\n\n    $ git clone https://github.com/makenew/python-app.git\n    $ cd python-app\n    $ poetry install\n\nRun each command below in a separate terminal window:\n\n::\n\n    $ make watch\n    $ make server\n\nPrimary development tasks are defined in the `Makefile`.\n\nSource Code\n~~~~~~~~~~~\n\nThe `source code`_ is hosted on GitHub.\nClone the project with\n\n::\n\n    $ git clone https://github.com/makenew/python-app.git\n\n.. _source code: https://github.com/makenew/python-app\n\nRequirements\n~~~~~~~~~~~~\n\nYou will need `Python 3`_ and Poetry_.\n\nInstall the development dependencies with\n\n::\n\n    $ poetry install\n\n.. _Poetry: https://poetry.eustace.io/\n.. _Python 3: https://www.python.org/\n\nTests\n~~~~~\n\nLint code with\n\n::\n\n    $ make lint\n\n\nRun tests with\n\n::\n\n    $ make test\n\nRun tests on chages with\n\n::\n\n    $ make watch\n\nPublishing\n~~~~~~~~~~\n\nUse the bump2version_ command to release a new version.\nPush the created git tag which will trigger a GitHub action.\n\n.. _bump2version: https://github.com/c4urself/bump2version\n\nPublishing may be triggered using on the web\nusing a `workflow_dispatch on GitHub Actions`_.\n\n.. _workflow_dispatch on GitHub Actions: https://github.com/makenew/python-app/actions?query=workflow%3Aversion\n\nGitHub Actions\n--------------\n\n*GitHub Actions should already be configured: this section is for reference only.*\n\nThe following repository secrets must be set on GitHub Actions.\n\n- ``PYPI_API_TOKEN``: API token for publishing on PyPI.\n- ``GH_USER``: The GitHub user\'s username.\n- ``GH_TOKEN``: A personal access token for the user.\n\nThese must be set manually.\n\nSecrets for Optional GitHub Actions\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nThe version and format GitHub actions\nrequire a user with write access to the repository\nincluding access to read and write packages.\nSet these additional secrets to enable the action:\n\n- ``GH_USER``: The GitHub user\'s username.\n- ``GH_TOKEN``: A personal access token for the user.\n- ``GIT_USER_NAME``: The name to set for Git commits.\n- ``GIT_USER_EMAIL``: The email to set for Git commits.\n- ``GPG_PRIVATE_KEY``: The `GPG private key`_.\n- ``GPG_PASSPHRASE``: The GPG key passphrase.\n\n.. _GPG private key: https://github.com/marketplace/actions/import-gpg#prerequisites\n\nContributing\n------------\n\nPlease submit and comment on bug reports and feature requests.\n\nTo submit a patch:\n\n1. Fork it (https://github.com/makenew/python-app/fork).\n2. Create your feature branch (`git checkout -b my-new-feature`).\n3. Make changes.\n4. Commit your changes (`git commit -am \'Add some feature\'`).\n5. Push to the branch (`git push origin my-new-feature`).\n6. Create a new Pull Request.\n\nLicense\n-------\n\nThis Python app is licensed under the MIT license.\n\nWarranty\n--------\n\nThis software is provided by the copyright holders and contributors "as is" and\nany express or implied warranties, including, but not limited to, the implied\nwarranties of merchantability and fitness for a particular purpose are\ndisclaimed. In no event shall the copyright holder or contributors be liable for\nany direct, indirect, incidental, special, exemplary, or consequential damages\n(including, but not limited to, procurement of substitute goods or services;\nloss of use, data, or profits; or business interruption) however caused and on\nany theory of liability, whether in contract, strict liability, or tort\n(including negligence or otherwise) arising in any way out of the use of this\nsoftware, even if advised of the possibility of such damage.\n',
    'author': 'Evan Sosenko',
    'author_email': 'razorx@evansosenko.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/makenew/python-app',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.9.7',
}


setup(**setup_kwargs)
